/* The Last Ninja Remixed
 *  
 * Copyright (C) 2008-2009 Robert Crossfield
 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * 
 * $Id: surface.cpp 12 2009-04-10 04:08:13Z segra $
 * 
 */

#include "..\stdafx.h"
#include "..\scale\scalebit.h"

namespace NinjaRemixed {

	graphicSurface::graphicSurface(	word width, word height, const byte *buffer, size_t bufferSize ) {

		// Clear some things
		_buffer			= buffer;
		_bufferSize		= bufferSize;
		_loaded			= false;
		
		_maxWidth		= width;
		_maxHeight		= height;

		_scaled			= false;

		// Get ourselves a surface
		_surface = SDL_CreateRGBSurface(	SDL_SWSURFACE,	width,	height,	 32, 0, 0, 0, 0);

		// Set the transparency key
		SDL_SetColorKey(	_surface, SDL_SRCCOLORKEY, SDL_MapRGB(_surface->format, 0xFF, 0xFF, 0xFF)	);

	}

	graphicSurface::~graphicSurface() {

		if( _surface )
			SDL_FreeSurface( _surface );

	}

	void graphicSurface::blit( graphicSurface	*Source, dword mask ) {
		if(!Source)
			return;

		dword  maskCurrent = mask;
		dword *bufferSrc = Source->pixelGet( 0, 0 );
		dword *bufferDst = pixelGet( 0, 0 );
		word x, y, maxX, maxY;

		maxX = Source->widthGet();
		maxY = Source->heightGet() - 1;

		for(y = 0; y < maxY; y++) {
			for(x = 0; x < maxX; x++) {

				if(*bufferSrc && (maskCurrent & 0x80000000))
					*bufferDst++ = *bufferSrc++;
					
				else {
					bufferSrc++;
					bufferDst++;
				}
				
				maskCurrent <<= 1;
				if(maskCurrent == 0)
					maskCurrent = mask;
			}
		}
	}

	void graphicSurface::blit( graphicSurface *Source ) {
		blit( Source, 0, 0, 0xFF );
	}

	void graphicSurface::unblit( graphicSurface	*Source, word destX, word destY, dword colorKey ) {
		if(!Source)
			return;
			
		if( Source->scaleGet() ) {
			destX *= scaleLevel;
			destY *= scaleLevel;
		}

		dword *bufferSrc = Source->pixelGet( 0, 0 );
		dword *bufferDst = pixelGet( destX, destY );
		word x, y, maxX, maxY;

		maxX = Source->widthGet();
		maxY = Source->heightGet(); // -1

		dword difference = _maxWidth - maxX;

		for(y = 0; y < maxY; y++) {
			for(x = 0; x < maxX; x++) {
	
				if(*bufferSrc == *bufferDst && *bufferSrc != colorKey ) {
					*bufferDst++ = 0x00;
					bufferSrc++;
				} else {
					bufferSrc++;
					bufferDst++;
				}

			}

			bufferDst += difference;
		}
	}

	void graphicSurface::blit( graphicSurface	*Source, word destX, word destY, dword colorKey, dword colorKey2 ) {
		if(!Source)
			return;
			
		if( Source->scaleGet() ) {
			destX *= scaleLevel;
			destY *= scaleLevel;
		}

		dword *bufferSrc = Source->pixelGet( 0, 0 );
		dword *bufferDst = pixelGet( destX, destY );
		word x, y, maxX, maxY;

		maxX = Source->widthGet();
		maxY = Source->heightGet(); // -1

		dword difference = _maxWidth - maxX;

		for(y = 0; y < maxY; y++) {
			for(x = 0; x < maxX; x++) {
	
				if(*bufferSrc != colorKey && *bufferSrc != colorKey2 )
					*bufferDst++ = *bufferSrc++;
					
				else {
					bufferSrc++;
					bufferDst++;
				}

			}

			bufferDst += difference;
		}
	}

	void graphicSurface::draw ( word X, word Y, bool scale ) {
		byte	*bufferCurrent	= (byte*) _buffer;

		if(!_loaded && bufferCurrent) {
			// Load the image
			load( _buffer, _bufferSize );
			
			bufferCurrent = (byte*) _buffer;
		}

		clear();

		while( bufferCurrent < ( byte*) (_buffer + _bufferSize ) ) {
			
			pixelDraw( X, Y,  *bufferCurrent );

			bufferCurrent++;

			// Increase X then check if its reached the screen edge
			if(++X >= _maxWidth) {
				X = 0;

				// Check the height
				if(++Y >= (_maxHeight)) 
					break;
			} 

		}	// while( bufferCurrent < ( byte*) (_buffer + _bufferSize ) )

		if(scale)
			// Scale it up
			scaleTo();
	}

	void graphicSurface::draw( bool scale ) {
		draw( 0, 0, scale );
	}

	void graphicSurface::clear( word X, word Y, word sizeX, word sizeY, byte color ) {
		SDL_Rect dest;

		dest.x = X;		dest.y = Y; 
		dest.w = sizeX; dest.h = sizeY; 

		SDL_FillRect( _surface, &dest, color );
	}

	void graphicSurface::clear( byte color ) {
		
		SDL_FillRect( _surface, NULL, color );
	}

	void graphicSurface::decode(	size_t numBits, byte colorMask, const byte *buffer, size_t bufferSize ) {
		byte	 bitCount, paletteIndex, count, modifier;
		word	*bufferCurrent = (word*) buffer;
		byte	*bufferDest;
		dword	 colorIndex[4];
		
		// Set our buffer size   ( 8 size of dword, pixel size )
		_bufferSize = ( bufferSize / numBits ) * 8;

		// Create the buffer
		_buffer = new byte[ _bufferSize ];
		bufferDest = (byte*) _buffer;

		clear();

		while( bufferCurrent < (word*) (buffer + bufferSize) ) {

			for( count = 0; count < numBits; count++)
				colorIndex[count] = swapDWord( readWord( bufferCurrent++ ) );

			for( bitCount = 0; bitCount < 16; bitCount++) {
				paletteIndex = 0x00;
					
				// Read the index bit for the pixel from each bitplane
				for( count = 0, modifier = 1; count < numBits; count++ ) {
					
					if( (colorIndex[count] & 0x80000000))
						paletteIndex += modifier;

					modifier *= 2;

					// Shift the pixels
					colorIndex[count] <<= 1;
				}

				if(paletteIndex == 0x00)
					paletteIndex =  colorMask;

				// Draw the color
				(*bufferDest++) = paletteIndex;

			}	// for bitCount < 16

		}	// while bufferCurrent

		_loaded = true;	
	}

	void graphicSurface::paletteLoad(		const byte  *buffer	) {
		word  color;
		byte  colorID, colorRed, colorGreen, colorBlue;

		for( colorID = 0; colorID < colorsMax; colorID++) {
			
			// Get the next color codes
			color = readWord( buffer );
			buffer+=2;

			color *= 2;

			// Extract each color from the word
			//  X X X X   R3 R2 R1 R0     G3 G2 G1 G0   B3 B2 B1 B0
			colorRed	=	(color >> 8) & 0xF;
			colorGreen	=	(color >> 4) & 0xF;
			colorBlue	=	(color >> 0) & 0xF;
			
			colorRed *= 2;	colorGreen *= 2;	colorBlue *= 2;
			colorRed *= 2;	colorGreen *= 2;	colorBlue *= 2;
			colorRed *= 2;	colorGreen *= 2;	colorBlue *= 2;
			colorRed *= 2;	colorGreen *= 2;	colorBlue *= 2;

			// Set the color in the palette
			paletteColorSet( colorID, colorRed, colorGreen, colorBlue );
		}

	}

	inline paletteColor	graphicSurface::paletteColorGet(	byte id ) {
		// If its outside the range, return the first color
		if(id >= colorsMax)
			id = 0;

		return _palette[id];
	}

	inline void graphicSurface::paletteColorSet( byte id, byte red, byte green, byte blue ) {
		if(id >= colorsMax)
			return;

		_palette[id].red	= red;
		_palette[id].green	= green;
		_palette[id].blue	= blue;
	}

	inline dword	*graphicSurface::pixelGet( word x,	word y ) {
		dword	*_pixelPosition;

		// Adjust to beginning of screen
		_pixelPosition = (dword*) _surface->pixels ;
		
		// Offset by Y
		_pixelPosition +=  (y * (_surface->pitch / 4) );

		// Offset by X
		_pixelPosition +=  x;

		return _pixelPosition;
	}

	void graphicSurface::pixelDraw(			word x,		word y,		dword paletteIndex ) {
		dword	*_pixelPosition;

		// Get the color from the palette
		dword color = SDL_MapRGB (	_surface->format , _palette[paletteIndex].red , _palette[paletteIndex].green , _palette[paletteIndex].blue ) ;

		_pixelPosition = pixelGet( x, y );

		// set pixel data
		*_pixelPosition =  color;

	}

	SDL_Surface *graphicSurface::surfaceGet( ) {
		return _surface;
	}

	void graphicSurface::scaleTo() {
		bool scaled = _scaled;

		SDL_Surface *tmpScaled = 0;

		// Invalid Scale?
		if(scaleLevel < 2 || scaleLevel > 4)
			return;

		if(_scaled == false) {
			// Set our new dimensions
			_maxWidth	= _surface->w * scaleLevel;
			_maxHeight	= _surface->h * scaleLevel;
			_scaled = true;	
		}

		// Create a new surface
		tmpScaled = SDL_CreateRGBSurface(SDL_SWSURFACE, _maxWidth, _maxHeight, 32, 0, 0, 0, 0);

		SDL_SetColorKey(	tmpScaled, SDL_SRCCOLORKEY, SDL_MapRGB(tmpScaled->format, 0xFF, 0xFF, 0xFF)	);

		// Do the scale
		if(scaled)
			scale(scaleLevel, tmpScaled->pixels, tmpScaled->pitch, _surface->pixels, _surface->pitch, _surface->format->BytesPerPixel, _surface->w / scaleLevel, _surface->h / scaleLevel);
		else
			scale(scaleLevel, tmpScaled->pixels, tmpScaled->pitch, _surface->pixels, _surface->pitch, _surface->format->BytesPerPixel, _surface->w, _surface->h);

		// Free the old surface
		SDL_FreeSurface( _surface );

		_surface = tmpScaled;
	}

	word graphicSurface::heightGet() {
		return _maxHeight;
	}

	word graphicSurface::widthGet() {
		return _maxWidth;
	}
}	// NinjaRemixed
